#include "3ds.hpp"
#include <cstring>
#include <stdlib.h>
#include <cstdio>
#include <tremor/ivorbisfile.h>

#ifdef BUILD_SONIC_1
#include "splash_sonic1.h"
#else
#include "splash_sonic2.h"
#endif

static Sint16 clearTimer = 0;

//system
void n3ds_Init()
{
	srand(time(0));
	osSetSpeedupEnable(true);
	cfguInit();

	gfxInitDefault();
	consoleInit(GFX_BOTTOM, NULL);
	gfxSetWide(false);

	//display splash screen
	memcpy(gfxGetFramebuffer(GFX_TOP, GFX_LEFT, NULL, NULL), SPLASH_DATA, SPLASH_SIZE);
	gfxFlushBuffers();
	gfxSwapBuffers();
}

void n3ds_Exit()
{

}

bool n3ds_MainLoop()
{
	if (clearTimer > 0) {
		clearTimer -= 1;
		if (clearTimer <= 0) {
			clearTimer = 0;
			consoleClear();
		}
	}

	return aptMainLoop();
}


void* n3ds_CreateThread(void (*func)(void*), void* arg, int core)
{
	s32 prio = 0;
	svcGetThreadPriority(&prio, CUR_THREAD_HANDLE);
	return (void*)threadCreate(func, arg, 4*1024, prio-1, core, true);
}

int n3ds_GetLanguage()
{
	u8 l;
	CFGU_GetSystemLanguage(&l);

	int ret = 0;

	switch (l)
	{
		case CFG_LANGUAGE_EN: ret = 0; break;
		case CFG_LANGUAGE_FR: ret = 1; break;
		case CFG_LANGUAGE_IT: ret = 2; break;
		case CFG_LANGUAGE_DE: ret = 3; break;
		case CFG_LANGUAGE_ES: ret = 4; break;
		//case CFG_LANGUAGE_JP: ret = 5; break;
		case CFG_LANGUAGE_PT: ret = 6; break;
		//case CFG_LANGUAGE_RU: ret = 7; break;
		//case CFG_LANGUAGE_KO: ret = 8; break;
		//case CFG_LANGUAGE_ZH: ret = 9; break;
		//case CFG_LANGUAGE_TW: ret = 10; break;
		default: ret = 0;
	};

	return ret;
}

void n3ds_Print(const char *msg, ...)
{
	char buf[0x200];

    va_list args;
    va_start(args, msg);
    vsprintf(buf, msg, args);

	printf("%s", buf);
	clearTimer = 60*3;
}

void n3ds_Warning(const char *msg, ...)
{
	char buf[0x200];
	
    va_list args;
    va_start(args, msg);
    vsprintf(buf, msg, args);

	consoleClear();
	printf("\n!!! WARNING !!!\n\n");
	printf("%s\n", buf);
	printf("\n(A) - Continue\n");

	while (aptMainLoop()) {
		hidScanInput();
		if (hidKeysHeld() & KEY_A)
			break;
	}

	consoleClear();
}

//video
static int screen_w = 400;
static int screen_h = 240;
static int xoffset = 0;
static int game_w = 400;
static gfxScreen_t screen = GFX_TOP;

void n3ds_InitGraphics()
{
	gfxSetScreenFormat(screen, GSP_RGB565_OES);
}

void n3ds_ExitGraphics()
{
	gfxSetWide(false);
	gfxExit();
}

void n3ds_SetScreenWidth(int xsize)
{
	if (xsize > 400) {
		//800px mode
		gfxSetWide(true);
		screen_w = 800;
		screen = GFX_TOP;
	}
	else {
		//regular mode
		gfxSetWide(false);

		if (xsize > 320) {
			//top screen
			screen_w = 400;
			screen = GFX_TOP;
		}
		else {
			//bottom screen
			screen_w = 320;
			screen = GFX_BOTTOM;
			consoleClear();
			consoleInit(GFX_TOP, NULL);
		}
	}

	xoffset = (screen_w - xsize) / 2;
	game_w = xsize;

	gfxSetScreenFormat(screen, GSP_RGB565_OES);
	//clear garbage from borders
	gfxSwapBuffers();
	u8* fb = gfxGetFramebuffer(screen, GFX_LEFT, 0, 0);
	memset(fb, 0, screen_w*screen_h*2);
}

void n3ds_CopyFrameBuffer(Uint16* pixels, bool vsync)
{
	Uint16* fb = (Uint16*)gfxGetFramebuffer(screen, GFX_LEFT, 0, 0);

	for (int y = 0; y < screen_h; y++) {
		for (int x = 0; x < game_w; x++) {
			fb[((x+xoffset) * screen_h) + (screen_h-y-1)] = *pixels++;
		}
	}

	gfxSwapBuffers();
	if (vsync)
		gspWaitForVBlank();
}

//audio
#pragma pack(push,1)
typedef struct
{
	unsigned int ChunkID;
	unsigned int ChunkSize;
	unsigned int Format;
	unsigned int Subchunk1ID;
	unsigned int Subchunk1Size;
	unsigned short AudioFormat;
	unsigned short NumChannels;
	unsigned int SampleRate;
	unsigned int ByteRate;
	unsigned short BlockAlign;
	unsigned short BitsPerSample;
	unsigned int Subchunk2ID;
	unsigned int Subchunk2Size;
} WAVHeader;
#pragma pack(push,0)

#define BUF_SIZE (8192)
#define NDSP_NUM_CHANNELS 2
static ndspWaveBuf waveBuf[NDSP_NUM_CHANNELS];
static volatile bool audioEnabled = false;
static volatile bool audioLock = false;
static volatile unsigned char fillBlock = 0;
static void (*processAudio) (void*, uint8_t*, int) = NULL;
static Thread threadID;
static Handle threadRequest;

static void _n3ds_AudioCallback(void *const arg)
{
	if (!audioEnabled)
		return;

	svcSignalEvent(threadRequest);
}

static void _n3ds_AudioThread(void* arg)
{
	ndspChnReset(0);
	ndspChnSetInterp(0, NDSP_INTERP_NONE);
	ndspChnSetRate(0, 44100);
	ndspChnSetFormat(0, NDSP_FORMAT_STEREO_PCM16);

	for (int i = 0; i < NDSP_NUM_CHANNELS; i++) {
		waveBuf[i].data_vaddr = (uint8_t*)linearAlloc(BUF_SIZE);
		waveBuf[i].nsamples = BUF_SIZE / 2 / 2;
		waveBuf[i].looping = false;
		waveBuf[i].status = NDSP_WBUF_DONE;
	}

	svcCreateEvent(&threadRequest, RESET_ONESHOT);
	ndspSetCallback(_n3ds_AudioCallback, NULL);

	while (audioEnabled)
	{
		if (audioLock == false && waveBuf[fillBlock].status == NDSP_WBUF_DONE)
		{
			processAudio(NULL, (uint8_t*)waveBuf[fillBlock].data_vaddr, BUF_SIZE);

			ndspChnWaveBufAdd(0, &waveBuf[fillBlock]);
			DSP_FlushDataCache(waveBuf[fillBlock].data_vaddr, BUF_SIZE);

			fillBlock = !fillBlock;
		}

		svcWaitSynchronization(threadRequest, UINT64_MAX);
    	svcClearEvent(threadRequest);
	}

	svcCloseHandle(threadRequest);

	for (int i = 0; i < NDSP_NUM_CHANNELS; i++)
		linearFree((uint8_t*)waveBuf[i].data_vaddr);
	audioEnabled = false;
}

int n3ds_InitAudio()
{
	if (ndspInit()) {
		char dsppath[] = "sdmc:/3ds/dspfirm.cdc";
		FILE* f = fopen(dsppath, "rb");
		if (!f)
			n3ds_Warning("ndspInit() failed.\n'%s' was not found.\nThere will be no audio.\n", dsppath);
		else			
			n3ds_Warning("ndspInit() failed for an unknown reason.\nThere will be no audio.\n");
		fclose(f);

		return 0;
	}	

	return 1;
}

void n3ds_ExitAudio()
{
	if (audioEnabled) {
		audioEnabled = false;
		svcSignalEvent(threadRequest);
		threadJoin(threadID, U64_MAX);
		//Threads are freed automatically. Freeing them manually results in an illegal write and/or program freeze.
		//threadFree(threadID);
	}	

	ndspExit();
}

void n3ds_LockAudio()
{
	audioLock = true;
}

void n3ds_UnlockAudio()
{
	audioLock = false;
}

void n3ds_SetAudioProcess(void (*proc)(void*, uint8_t*, int))
{
	processAudio = proc;	
	audioEnabled = true;
	threadID = (Thread)n3ds_CreateThread(_n3ds_AudioThread, NULL, CORE_ANY);
	if (threadID == NULL) {
		n3ds_Print("Failed to create music thread!\n");
		audioEnabled = false;
	}
}

Sint16* n3ds_LoadWavMem(void* fileBuf, size_t size, size_t* len)
{
	Sint16* out = NULL;
	FILE* src = fmemopen(fileBuf, size, "rb");

	if (src == NULL)
		return NULL;

	//
	WAVHeader wh;
	fread(&wh, 1, sizeof(WAVHeader), src);

	if (wh.NumChannels == 1)
	{
		if (wh.BitsPerSample == 8)
		{
			out = (Sint16*)malloc(wh.Subchunk2Size * 2 * 2);
			*len = wh.Subchunk2Size * 2 * 2 / sizeof(Sint16);
			
			for (int i = 0; i < wh.Subchunk2Size; i++)
			{
				char b;
				fread(&b, 1, 1, src);
				out[i*2] = (b-128) << 8;
				out[i*2+1] = (b-128) << 8;
			}				
		}
		else if (wh.BitsPerSample == 16)
		{
			out = (Sint16*)malloc(wh.Subchunk2Size * 2);
			*len = wh.Subchunk2Size * 2 / sizeof(Sint16);
			
			for (int i = 0; i < wh.Subchunk2Size/sizeof(Sint16); i++)
			{
				Sint16 b;
				fread(&b, sizeof(Sint16), 1, src);
				out[i*2] = b;
				out[i*2+1] = b;
			}
		}
	}
	else if (wh.NumChannels == 2)
	{
		if (wh.BitsPerSample == 8)
		{
			printf("2 channel 8-bit wavs currently unimplemented\n");
		}
		else if (wh.BitsPerSample == 16)
		{
			out = (Sint16*)malloc(wh.Subchunk2Size);
			*len = wh.Subchunk2Size / sizeof(Sint16);
			fread(out, 1, wh.Subchunk2Size, src);
		}
	}

	fclose(src);
	return out;
}

Sint16* n3ds_LoadOggMem(void* fileBuf, size_t size, size_t* len)
{
	Sint16* out = NULL;

	OggVorbis_File vf;
    int error = ov_open(fmemopen(fileBuf, size, "rb"), &vf, NULL, 0);
    if (error != 0) {
        ov_clear(&vf);
        printf("n3ds_LoadOggMem - ov_open failed\n");
        return NULL;
    }

    vorbis_info *vinfo;
    vinfo = ov_info(&vf, -1);
    int   channels = vinfo->channels;
    long  samples  = (long)ov_pcm_total(&vf, -1);
    uint  audioLen = (Uint32)(samples * channels * 2);
    char* audioBuf = (char*)malloc(audioLen);
    
    char* buf = audioBuf;
    long toRead = audioLen;
    long read = 0;
    int bitstream = -1;

    do {
        read = ov_read(&vf, buf, toRead, &bitstream);

        buf    += read;
        toRead -= read;
    }
    while (read > 0);
    
    ov_clear(&vf);
    
    if (channels == 1) {
    	//convert mono to dual mono
        Sint16* convert = (Sint16*)malloc(audioLen * sizeof(Sint16));
        Sint16* in = (Sint16*)audioBuf;
        for (int i = 0; i < audioLen / sizeof(Sint16); i++) {
            convert[i*2]   = in[i];
            convert[i*2+1] = in[i];
        }
        free(audioBuf);

         out = convert;
        *len = audioLen * 2 / sizeof(Sint16);
    }
    else {
         out = (Sint16*)audioBuf;
        *len = audioLen / sizeof(Sint16);
    }

    return out;
}

//input
static u32 kDown = 0;

static u32 buttons[] = {
	KEY_A, 		//0
	KEY_B,		//1
	KEY_Y,		//2
	KEY_X,		//3
	KEY_SELECT,	//4
	KEY_TOUCH,	//5
	KEY_START,	//6
	0,			//7
	0,			//8
	KEY_L,		//9
	KEY_R,		//10
	KEY_UP,		//11
	KEY_DOWN,	//12
	KEY_LEFT,	//13
	KEY_RIGHT,	//14
	0,			//15
	KEY_ZL,		//16
	KEY_ZR		//17
};

void n3ds_ScanInput()
{
	hidScanInput();
	kDown = hidKeysHeld();
}

bool n3ds_GetButtonDown(int btn)
{
	if (btn < 0 || btn > 17)
		return false;

	return (kDown & buttons[btn]);
}